"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.api = void 0;
const chai_1 = require("chai");
const client_1 = require("@deep-foundation/hasura/client");
const client_2 = require("../imports/client");
const handlers_1 = require("../imports/handlers");
const api_1 = require("@deep-foundation/hasura/api");
const sql_1 = require("@deep-foundation/hasura/sql");
const debug_1 = __importDefault(require("debug"));
const debug = (0, debug_1.default)('deeplinks:tests:in-transaction-handlers');
const log = debug.extend('log');
const error = debug.extend('error');
const namespaces = debug_1.default.disable();
debug_1.default.enable(`${namespaces ? `${namespaces},` : ``}${error.namespace}`);
const HASURA_PATH = 'localhost:8080';
const HASURA_SSL = 0;
const HASURA_SECRET = 'myadminsecretkey';
exports.api = new api_1.HasuraApi({
    path: HASURA_PATH,
    ssl: !!+HASURA_SSL,
    secret: HASURA_SECRET,
});
const apolloClient = (0, client_1.generateApolloClient)({
    path: `${process.env.DEEPLINKS_HASURA_PATH}/v1/graphql`,
    ssl: !!+process.env.DEEPLINKS_HASURA_SSL,
    secret: process.env.DEEPLINKS_HASURA_SECRET,
});
const deep = new client_2.DeepClient({ apolloClient });
let lastHandlerLinkId = 99999999;
const nextHandlerResult = () => {
    lastHandlerLinkId -= 1;
    return lastHandlerLinkId;
};
describe('sync handlers', () => {
    describe('DeepClient mini', () => {
        it(`id`, () => __awaiter(void 0, void 0, void 0, function* () {
            var _a, _b, _c, _d, _e, _f;
            const result = yield exports.api.sql((0, sql_1.sql) `select links__deep__client('id', '{"start": "@deep-foundation/core", "path":["RuleSubject"]}'::json)`);
            const clientResult = yield deep.id('@deep-foundation/core', 'RuleSubject');
            log((_c = (_b = (_a = result === null || result === void 0 ? void 0 : result.data) === null || _a === void 0 ? void 0 : _a.result) === null || _b === void 0 ? void 0 : _b[1]) === null || _c === void 0 ? void 0 : _c[0], clientResult);
            chai_1.assert.equal((_f = (_e = (_d = result === null || result === void 0 ? void 0 : result.data) === null || _d === void 0 ? void 0 : _d.result) === null || _e === void 0 ? void 0 : _e[1]) === null || _f === void 0 ? void 0 : _f[0], clientResult);
        }));
    });
    describe('Handle operations', () => {
        it(`Handle insert`, () => __awaiter(void 0, void 0, void 0, function* () {
            var _a, _b, _c;
            const CustomNumber = nextHandlerResult();
            const typeId = yield deep.id('@deep-foundation/core', 'Type');
            const handleInsertTypeId = yield deep.id('@deep-foundation/core', 'HandleInsert');
            const supportsId = yield deep.id('@deep-foundation/core', 'plv8SupportsJs');
            const handler = yield (0, handlers_1.insertHandler)(handleInsertTypeId, typeId, `(deep, data) => { deep.insert({id: ${CustomNumber}, type_id: 2}); }`, undefined, supportsId);
            log('handler', handler);
            log('CustomNumber', CustomNumber);
            const linkId = yield (0, handlers_1.ensureLinkIsCreated)(typeId);
            log('linkId', linkId);
            const insertedByHandler = (yield deep.select({ id: { _eq: CustomNumber } }));
            log('insertedByHandler', (_a = insertedByHandler === null || insertedByHandler === void 0 ? void 0 : insertedByHandler.data[0]) === null || _a === void 0 ? void 0 : _a.id);
            if ((_b = insertedByHandler === null || insertedByHandler === void 0 ? void 0 : insertedByHandler.data[0]) === null || _b === void 0 ? void 0 : _b.id)
                yield deep.delete({ id: { _eq: CustomNumber } });
            log('delete linkid', yield deep.delete({ id: { _eq: linkId } }));
            log('delete handler', yield (0, handlers_1.deleteHandler)(handler));
            chai_1.assert.equal((_c = insertedByHandler === null || insertedByHandler === void 0 ? void 0 : insertedByHandler.data[0]) === null || _c === void 0 ? void 0 : _c.id, CustomNumber);
        }));
        it.skip(`Handle insert on selector`, () => __awaiter(void 0, void 0, void 0, function* () {
            const CustomNumber = nextHandlerResult();
            const handleInsertTypeId = yield deep.id('@deep-foundation/core', 'HandleInsert');
            const selector = yield (0, handlers_1.insertSelector)();
            const { nodeTypeId, linkTypeId, treeId, selectorId, rootId } = selector;
            const handler = yield (0, handlers_1.insertHandler)(handleInsertTypeId, selectorId, `(arg) => {
          plv8.execute("INSERT INTO strings (link_id, value) VALUES(${rootId}, 'test')");
        }`);
            const selectorItems = yield (0, handlers_1.insertSelectorItem)({ selectorId, nodeTypeId, linkTypeId, treeId, rootId });
            log('selectorItems', selectorItems);
            for (const selectorItem of selectorItems) {
                yield (0, handlers_1.deleteId)(selectorItem.linkId);
                yield (0, handlers_1.deleteId)(selectorItem.nodeId);
            }
            yield (0, handlers_1.deleteSelector)(selector);
            yield (0, handlers_1.deleteHandler)(handler);
        }));
    });
});
//# sourceMappingURL=in-transaction-handlers.js.map